"""
Tools to help with performing data analysis.

These were originally built specifically for managing and interacting with FID
gas chromatograms. Future development should treat data sets as ambiguously as
possible so any data types (e.g. mass spec, FTIR) can utilize the same tools.
Created on Mon Feb  6 11:46:49 2023.
@author: Briley Bourgeois
"""
import os
import pickle
import re
from math import sqrt

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from catalight.analysis.gcdata import GCData
from catalight.equipment.experiment_control import Experiment


def list_matching_files(main_dirs, target, suffix):
    """
    Return files containing 'target' and ending in 'suffix' within a folder.

    Returns the full path of all files within the give folder path that contain
    the phrase 'target' and ends with 'suffix'. This function crawls the ENTIRE
    directory provided, including all sub-folders down to bottom depth.

    Parameters
    ----------
    main_dirs : list of str
        List of string paths to main directories to crawl for data.
        Can be a list of one to search a single directory.
    target : str
        String of target phrase for finding files.
    suffix : str
        String target for the file type to search for.

    Returns
    -------
    list(str)
        List of filepaths ending with 'suffix' and containing 'target'
    """
    if isinstance(main_dirs, str):
        main_dirs = [main_dirs]  # If user supplies str, convert to 1 item list
    filepath_list = []
    for root in main_dirs:
        for dirpath, dirnames, filenames in os.walk(root):
            for filename in filenames:
                if (filename.lower().endswith(suffix.lower())
                        & (target.lower() in filename.lower())):
                    filepath = os.path.join(dirpath, filename)
                    filepath_list.append(filepath)
    return filepath_list


def list_expt_obj(file_paths):
    """
    Get list of experiments.

    Parameters
    ----------
    file_paths : list[str]
        List of full paths to expt_log.txt files.

    Returns
    -------
    list[Experiment]
        List of Experiment objects from provided filepaths.
    """
    experiments = []
    # If user enters one filepath as str, convert to list for them.
    if isinstance(file_paths, str):
        file_paths = [file_paths]

    for log_path in file_paths:
        # Double check filename is correct
        if os.path.basename(log_path) == 'expt_log.txt':
            expt_path = os.path.dirname(log_path)
            expt = Experiment()  # Initialize experiment obj
            expt.read_expt_log(log_path)  # Read expt parameters from log
            expt.update_save_paths(expt_path)  # update file paths
            experiments.append(expt)
        else:
            print('Path other than expt_log.txt passed to list_expt_obj()')
    return experiments


def build_results_dict(file_list, data_labels, reactant, target):
    """
    Generate a results dictionary of X/S and data labels for paths given.

    Results dict contains [X, S, X Error, S Error] for given datasets and
    labels typically for use with plotting.

    Parameters
    ----------
    file_list : list of str
        List of file paths to experiment folders or experiment log paths then
        used to initiate experiment objects and calculate conversion and
        selectivity.
    data_labels : list of str
        List of data labels used for generating plot legends.
    reactant : str
        String identity of reactant molecule to track. Must match what exists
        in the calibration file exactly.
    target : str
        String identity of the target to use when calculating selectivity. Must
        match what exists in the calibration file exactly.

    Returns
    -------
    dict :
        Dictionary of results (produced by calculate_X_and_S) in the form
        {data label: result}
        where data label is a string with the desired legend label and
        result is a pandas.DataFrame with column headers
        ['Conversion', 'Selectivity', 'X Error', 'S Error']
        This function is most easily used in conjunction with DataExtractor
        such as in the run_plot_comparison script.

    """
    results_dict = {}
    # If the file_list is expt folders, convert to experiment log paths
    if (not file_list[0].endswith('expt_log.txt')
            and os.path.isfile(os.path.join(file_list[0], 'expt_log.txt'))):
        for n in range(len(file_list)):
            file_list[n] = os.path.join(file_list[n], 'expt_log.txt')
    elif file_list[0].endswith('expt_log.txt'):
        # At least first file is ok to move on
        pass
    else:
        msg = 'Item 1 of file_list is not experiment log or experiment folder'
        raise AttributeError(msg)

    expt_list = list_expt_obj(file_list)
    for expt, data_label in zip(expt_list, data_labels):
        result = calculate_X_and_S(expt, reactant, target)
        results_dict[data_label] = result
    return results_dict


def get_run_number(filename):
    """
    Return run number based on filename.

    Parameters
    ----------
    filename : str
        string possibly containing number ID at end

    Returns
    -------
    int or None
        returns number suffix if present from file name.

    """
    parts = filename.split('.')  # Separate filename at each period
    # Second to last part always has run num
    run_num_part = parts[len(parts) - 2]
    # \d digit; + however many; $ at end
    run_number = re.search(r'\d+$', run_num_part)
    return int(run_number.group()) if run_number else None


def get_bool(prompt):
    """Convert common strings to bool."""
    while True:
        acceptable_answers = {"true": True,
                              "false": False,
                              'yes': True,
                              'no': False}
        try:
            return acceptable_answers[input(prompt).lower()]
        except KeyError:
            print("Invalid input please enter True or False!")


def load_results(expt):
    """
    Load analysis results of previously analyzed experiment.

    Parameters
    ----------
    expt : Experiment
        Experiment object that has data and has been analyzed.

    Returns
    -------
    numpy.ndarray:
        3D matrix of concentrations for each molecule, gc collection,
        and condition
        [Condition x [Timestamps, ChemID] x run number
    pandas.DataFrame:
        average concentration for each molecule and experiment condition
    pandas.DataFrame:
        one standard deviation of concentration measurements
    """
    fol = expt.results_path
    avg = pd.read_csv(os.path.join(fol, 'avg_conc.csv'), index_col=(0))
    std = pd.read_csv(os.path.join(fol, 'std_conc.csv'), index_col=(0))
    concentrations = np.load(os.path.join(fol, 'concentrations.npy'))
    return (concentrations, avg, std)


def convert_index(dataframe):
    """
    Take in dataframe, convert index from string to float.

    If overrun data exists, drops those rows.

    Parameters
    ----------
    dataframe : pandas.DataFrame
        Pandas dataframe containing index with string labels

    Returns
    -------
    pandas.DataFrame
        Same Pandas dataframe, now containing index with float labels
    """
    # scrub letters from first entry - old
    # unit = re.sub(r"\d+", "", dataframe.index[0])

    # very old data collected before automated control can have overflow data
    dataframe.drop('Over_Run_Data', errors='ignore', inplace=True)
    # remove letter and convert num string to float
    x_data = dataframe.index.str.replace(r'\D', '', regex=True).astype(float)
    dataframe.index = x_data
    return dataframe


def get_timepassed(concentrations, switch_to_hours=2, expt=None):
    """
    Convert time stamps to array of cumulative time passed.

    Parameters
    ----------
    concentrations : numpy.ndarray
        3D matrix of concentrations for each molecule, gc collection,
        and condition
        [Condition x [Timestamps, ChemID] x run number
    switch_to_hours : float, optional
        Time in hours when the output should switch units to
        hours instead of minutes. The default is 2.

    Returns
    -------
    numpy.ndarray :
        Numpy array of cumulative time passed since the start of experiment.
    str :
        Either 'min' or 'hr' based on the length of total time and parameters.

    """
    time_stamps = concentrations[:, 0, :].reshape(-1)
    time_stamps = time_stamps[~np.isnan(time_stamps)]

    # Experiments after 20230225 save w/ timestamp
    if isinstance(expt.start_time, float):
        start_time = expt.start_time

    else:  # Method before 20230225 (less accurate)
        start_time = np.min(time_stamps)

    if np.max(time_stamps) > (switch_to_hours * 60 * 60):
        time_passed = (time_stamps - start_time) / 60 / 60
        time_unit = 'hr'
    else:
        time_passed = (time_stamps - start_time) / 60
        time_unit = 'min'

    return time_passed, time_unit


def analyze_cal_data(expt, calDF, figsize=(6.5, 4.5), force_zero=True):
    """
    Run analysis on previously collected calibration data.

    Calibration data should look like a composition sweep experiment and a
    calibration data frame containing the gasses to be analyzed, their
    respective starting concentrations, and the elution time from the gc needs
    to be provided.

    Parameters
    ----------
    expt : Experiment
        Experiment object for already run calibration experiment
    calDF : pandas.DataFrame
        Formatted DataFrame containing gc calibration data.
        Specific to control file used!
        Format [ChemID, slope, intercept, start, end, ppm]
    figsize : tuple, optional
        Desired size of output figure in inches (x,y), Default is (6.5, 4.5).
        figsize suggestions:
        1/2 slide = (6.5, 4.5);  1/6 slide = (4.35, 3.25);
        1/4 slide =  (5, 3.65); Full slide =    (9, 6.65);
    force_zero : bool, optional
        Add point (x=0, y=0) to data set. The default is True.

    Returns
    -------
    tuple(matplotlib.axes._axes.Axes, matplotlib.axes._axes.Axes)
        (run_num_plots, calibration_plots) Axis handles for "run_num" plot and
        for plots showing expected ppm vs measured counts

    """
    print('Analyzing calibration data')
    plt.close('all')

    # Returns raw counts with error bars for each chemical in each folder
    concentrations, avg, std = run_analysis(expt, calDF)
    # plotting.set_plot_style(figsize)
    plt.rcParams['figure.figsize'] = figsize
    plt.rcParams['svg.fonttype'] = 'none'

    # Calculations:
    calchemIDs = calDF.index.to_numpy()  # get chem IDs from calibration files
    new_calibration = calDF.copy()

    n_rows = len(calchemIDs) // 3
    n_cols = -(-len(calchemIDs) // n_rows)  # Gives Ceiling
    # Initialize run num plot
    fig_run_num, run_num_plots = plt.subplots(n_rows, n_cols)
    # Initialize ppm vs expected ppm plot
    fig_calibration, calibration_plots = plt.subplots(n_rows, n_cols)

    calgas_flow = np.array([])
    for condition in avg.index:
        calgas_flow = np.append(calgas_flow,
                                float(re.findall(r"([\d.]*\d+)" + 'CalGas',
                                                 condition, re.IGNORECASE)[0]))

    # Plotting:
    for chem_num in range(len(calchemIDs)):
        chemical = calchemIDs[chem_num]

        # +1 offsets from timestamps in row 0
        ind_concentrations = concentrations[:, chem_num + 1, :]
        ind_concentrations = ind_concentrations[~np.isnan(ind_concentrations)]

        ax_run_num = run_num_plots.ravel()[chem_num]  # flattens to 1d array
        ax_run_num.plot(ind_concentrations / 1000, 'o', label=chemical)
        ax_run_num.text(.5, .85, chemical,
                        horizontalalignment='center',
                        transform=ax_run_num.transAxes)

        ax_calibration = calibration_plots.ravel()[chem_num]
        expected_ppm = calDF.loc[chemical, 'ppm'] * calgas_flow
        ax_calibration.errorbar(avg[chemical] / 1000, expected_ppm,
                                xerr=std[chemical] / 1000, fmt='o')
        ax_calibration.ticklabel_format(axis='both', style='sci',
                                        scilimits=[-2, 2])

        # Fit w/ counts as 'y' and 'y_err'
        if force_zero:  # Add point (0, 0) w/ infinitesimal error
            x_data = np.append(0, expected_ppm)
            y_data = np.append(0, avg[chemical].to_numpy())
            y_err = np.append(1e-19, std[chemical].to_numpy())
        else:
            x_data = expected_ppm
            y_data = avg[chemical].to_numpy()
            y_err = std[chemical].to_numpy()
        try:
            p, V = np.polyfit(x_data, y_data, 1, cov=True, w=1 / y_err)
            m, b, err_m, err_b = (*p, np.sqrt(V[0][0]), np.sqrt(V[1][1]))

            # Convert linear fit of ppm vs counts
            # to linear fit of counts vs ppm (flip axes)
            m = 1 / m
            err_m = err_m
            # Propagate error using error propagation formula
            b = b / m
            err_b = sqrt(err_b**2 + b**2 * err_m**2) / m

            counts_fit = np.linspace(0, max(y_data), 100)
            ppm_fit = m * counts_fit + b

            # add fit to plot
            ax_calibration.plot(counts_fit / 1000, ppm_fit, '--r')
            fit_label = '\n'.join(["m: %4.2f\n+/- %4.2f" % (m, err_m),
                                   "b: %4.2f\n+/- %4.2f" % (b, err_b)])

            ax_calibration.text(.02, .5, fit_label,
                                horizontalalignment='left',
                                transform=ax_calibration.transAxes, fontsize=8)
            ax_calibration.text(1, 1.05, chemical,
                                horizontalalignment='right',
                                transform=ax_calibration.transAxes, fontsize=8)
        except (np.linalg.LinAlgError):
            label = chemical + '\nBad Fit'
            ax_calibration.text(.02, .75, label,
                                horizontalalignment='left',
                                transform=ax_calibration.transAxes, fontsize=8)
        new_calibration.loc[chemical, 'slope':'err_intercept'] = [m, err_m,
                                                                  b, err_b]

    ax_run_num = fig_run_num.add_subplot(111, frameon=False)
    # hide tick and tick label of the big axis
    ax_run_num.tick_params(labelcolor='none', which='both',
                           top=False, bottom=False, left=False, right=False)
    ax_run_num.set_xlabel("Run Number")
    ax_run_num.set_ylabel('Counts/1000')

    ax_calibration = fig_calibration.add_subplot(111, frameon=False)
    # hide tick and tick label of the big axis
    ax_calibration.tick_params(labelcolor='none', which='both',
                               top=False, bottom=False,
                               left=False, right=False)
    ax_calibration.set_ylabel("Expected ppm")
    ax_calibration.set_xlabel('Counts/1000')

    # Figure Export
    fig_run_num.tight_layout(pad=1, w_pad=0.4, h_pad=0.4)
    fig_calibration.tight_layout(pad=1, w_pad=0.4, h_pad=0.4)
    fig_run_num_path = os.path.join(expt.results_path,
                                    str(figsize[0])
                                    + 'w_run_num_plot_individuals')
    fig_calibration_path = os.path.join(expt.results_path,
                                        str(figsize[0])
                                        + 'w_calibration_plot')

    fig_run_num.savefig(fig_run_num_path + '.svg', format="svg")
    fig_calibration.savefig(fig_calibration_path + '.svg', format="svg")
    new_calibration.to_csv(os.path.join(expt.results_path, 'calibration.csv'))

    pickle.dump(fig_run_num, open(fig_run_num_path + '.pickle', 'wb'))
    pickle.dump(fig_calibration, open(fig_calibration_path + '.pickle', 'wb'))
    plt.show()
    print('Finished calibration analysis')
    return (run_num_plots, calibration_plots)


def run_analysis(expt, calDF, basecorrect='True', savedata='True'):
    """
    Compute the concentrations, averages, and error from GC runs.

    Sweeps gc data in an experiment data directory and computes ppm
    concentrations for every requested molecules in the .asc files. Returns
    a 3D matrix of concentrations, a 2D matrix of avg concentrations for each
    experimental condition, and a 2D matrix of errors (1 standard deviation)

    Parameters
    ----------
    expt : Experiment
        Experiment object for desired analysis.
    calDF : pandas.DataFrame
        Formatted DataFrame containing gc calibration data.
        Specific to control file used!
        Format [ChemID, slope, intercept, start, end]
    basecorrect : bool, optional
        Indicates whether or not to baseline correct individual gc data.
        The default is 'True'.
    savedata : bool, optional
        Indicates whether or not to save data. The default is 'True'.

    Returns
    -------
    numpy.ndarray
        3D matrix of concentrations for each molecule, gc collection,
        and condition
        [Condition x [Timestamps, ChemID] x run number]
    pandas.DataFrame
        average concentration for each molecule and experiment condition
    pandas.DataFrame
        one standard deviation of concentration measurements

    """
    # Analysis Loop
    # TODO add TCD part
    ###########################################################################
    print('Analyzing data...')
    print(expt.date + expt.expt_type + '_' + expt.expt_name)
    expt_data_fol = expt.data_path
    expt_results_fol = expt.results_path
    os.makedirs(expt_results_fol, exist_ok=True)  # Make dir if not there
    calchemIDs = calDF.index.to_numpy()  # get chem IDs from calibration files
    max_runs = 0
    step_path_list = []
    for dirpath, dirnames, filenames in os.walk(expt_data_fol):
        # only looks at .ASC
        num_data_points = len(list_matching_files(dirpath, 'FID', '.asc'))

        # determine bottom most dirs
        if not dirnames:
            step_path_list.append(dirpath)

        # Determines largest # of runs in any dir
        if num_data_points > max_runs:
            max_runs = num_data_points

    # Preallocate empty numpy array with NAN values
    num_fols = len(step_path_list)
    num_chems = int(len(calchemIDs))
    condition = np.full(num_fols, 0, dtype=object)
    concentrations = np.full((num_fols, num_chems + 1, max_runs), np.nan)
    # TODO create err np.array
    # err_concentrations = np.full((num_fols, num_chems + 1, max_runs), np.nan)

    # Loops through the ind var step and calculates conc in each data file
    for step_path in step_path_list:
        print(os.path.basename(step_path))
        step_num, step_val = os.path.basename(step_path).split(' ')
        step_num = int(step_num) - 1
        data_list = list_matching_files(step_path, 'FID', '.asc')
        condition[step_num] = step_val
        conc = []
        # conc_err = [] TODO
        for filepath in data_list:
            # data is an instance of a class, for signal use data.signal etc
            data = GCData(filepath, basecorrect=True)

            values = data.get_concentrations(calDF)
            # TODO add error output to GC_Data.get_concentrations()
            # values, err = data.get_concentrations(calDF)
            # conc_err.append(err.tolist())
            conc.append(values.tolist())

        num_runs = len(conc)
        # [Condition x [Timestamps, ChemID] x run number]
        concentrations[step_num, :, 0:num_runs] = np.asarray(conc).T
        # err_concentrations[step_num, :, 0:num_runs] = np.asarray(conc).T

    # Pull out "Active" Units from expt_list DF
    units = (expt.expt_list['Units']
             [expt.expt_list['Active Status']].to_string(index=False))

    if expt.expt_type == 'stability_test':
        # Reset "condition" to be time passed for stability tests
        # TODO This could check unit if expt_list updates
        # to have time instead of temp in future.

        # Don't switch unit, get time passed in minutes.
        time_passed, _ = get_timepassed(concentrations,
                                        switch_to_hours=1e9, expt=expt)
        # Make sure time is chronological
        order = np.argsort(time_passed)
        condition = time_passed[order]  # Rewrite condition as time passed

        idx_name = 'time [min]'  # We can get rid of this and else if TODO done
        avg_dat = concentrations[0, 1:, order]
        # TODO add err_concentration values to std
        std_dat = avg_dat * 0

    else:
        idx_name = (expt.ind_var + ' [' + units + ']')  # Sweep Parameter
        avg_dat = np.nanmean(concentrations[:, 1:, :], axis=2)
        # TODO add err_concentration values to std
        std_dat = np.nanstd(concentrations[:, 1:, :], axis=2)

    # Redefine condition list as pandas index so we can assign a name
    condition = pd.Index(condition, name=idx_name)

    if units == 'frac':  # Change label style to stacked compositions.
        condition = condition.str.replace('_', '\n')
        condition = condition.str.replace('frac', '')

    elif expt.expt_type == 'stability_test':
        pass  # Already Float

    else:  # Convert filenames to float w/o units.
        condition = condition.str.replace(r'\D', '', regex=True).astype(float)

    # Results
    ###########################################################################
    avg = pd.DataFrame(avg_dat, columns=calchemIDs, index=condition)
    # TODO add err_concentration values to std
    std = pd.DataFrame(std_dat, columns=calchemIDs, index=condition)
    if savedata:
        np.save(os.path.join(expt_results_fol, 'concentrations'),
                concentrations)
        avg.to_csv(os.path.join(expt_results_fol, 'avg_conc.csv'))
        std.to_csv(os.path.join(expt_results_fol, 'std_conc.csv'))
    print('Finished analyzing ' + expt.expt_name)
    return (concentrations, avg, std)


def calculate_X_and_S(expt, reactant, target_molecule):
    """
    Calculate conversion and selectivity for analyzed GC data.

    Parameters
    ----------
    expt : Experiment
        Experiment object which has data and has been analyzed
        (i.e. has avg_conc.csv etc.)
    reactant : str
        String identity of reactant molecule to track. Must match what exists
        in the calibration file exactly.
    target_molecule : str
        String identity of the target to use when calculating selectivity. Must
        match what exists in the calibration file exactly.

    Returns
    -------
    pandas.DataFrame
        DataFrame with column header ['Conversion', 'Selectivity', 'Error']
        Error is one standard deviation.

    Notes
    -----
    Conversion and error are calculated in fraction assuming a total mole
    balance of 1. Error calculations are based off the error propagation
    formula, resulting in the equations below:

    .. math::

      &X = 1 - \\frac{C_{reactant}}{C_{total}}

      &S = \\frac{C_{target}}{C_{total}*X}

      &\\sigma_{C_{tot}} = \\sqrt{\\sigma_{Molecule A}^{2} + ...
                                 + \\sigma_{Molecule N}^{2}}

      &\\sigma_{X} = \\sqrt{(\\frac{\\sigma_{C_{reactant}}} {C_{total}})^{2}
                            + (\\frac{\\sigma_{C_{total}}*C_{reactant}}
                                     {C_{total}^{2}})^{2}}

      &\\sigma_{S} = \\sqrt{(\\frac{\\sigma_{C_{target}}} {C_{total}*X})^{2}
                            + (\\frac{\\sigma_{C_{total}}*C_{target}}
                                     {C_{total}^{2}*X})^{2}
                            + (\\frac{\\sigma_{X}*C_{target}}
                                     {C_{total}*X^{2}})^{2}}

    """
    concentrations, avg, std = load_results(expt)
    # Compute relevant concentrations
    C_tot = avg.sum(axis=1)  # total conc. of all molecules
    C_reactant = avg[reactant]  # total conc. of reactant molecule
    C_tar = avg[target_molecule]  # total conc. of target molecule

    # Alternative method of calculating:
    # C_tot = concentrations[:,1:,:].sum(axis=1)
    # C_reactant = concentrations[:,1,:]
    # C_tar = concentrations[:, 2, :]
    # X = 1- C_reactant / C_tot
    # np.average(X, axis=1)
    # np.std(X, axis=1)
    # S = (C_tar / (C_tot * X))
    # np.average(S, axis=1)
    # np.std(S, axis=1)

    # Compute errors in concentrations
    # Error in total conc = quad sum of errors
    err_Ctot = np.sqrt((std**2).sum(axis=1))
    err_Cr = std[reactant]  # Error in reactant concentration
    err_Ctar = std[target_molecule]  # Error in target molecule concentration

    # Computes Conversion and Selectivity
    X = (1 - C_reactant / C_tot)  # conversion assuming mol bal of 1
    S = (C_tar / (C_tot * X))  # Selectivity
    S = S.fillna(0)

    # Compute total errors using error propagation formula
    X_err = np.sqrt((err_Cr/C_tot)**2
                    + (err_Ctot * C_reactant/C_tot**2)**2)
    # Simple Formula. Gives different numbers but should be same...
    # X_err_2 = X * np.sqrt((err_Ctot/C_tot)**2 + (err_Cr/C_reactant)**2)

    # Need check for is X is zero
    S_err = np.sqrt(
                    (err_Ctar / (C_tot * X))**2
                    + ((C_tar * err_Ctot) / (C_tot**2 * X))**2
                    + ((C_tar * X_err) / (C_tot / X**2))**2
                    )
    # Error in selectivity undefined when conversion is zero.
    X_err.replace([np.inf, -np.inf], 0, inplace=True)
    S_err.replace([np.inf, -np.inf], 0, inplace=True)
    X_err.fillna(0, inplace=True)
    S_err.fillna(0, inplace=True)
    # rel_err = std / avg
    # X_err = ((rel_err**2).sum(axis=1))**(1 / 2) * rel_err.max(axis=1)

    results = pd.concat([X, S, X_err, S_err], axis=1)*100
    results.columns = ['Conversion', 'Selectivity', 'X Error', 'S Error']
    return results



"""
Interface with SRI GC through Peaksimple interface via .dll file.

Created on Thu Oct 14 17:30:17 2021
@author: Briley Bourgeois
"""
import os
import re
import time

# python.NET is imported with the name clr (Common Language Runtime)
import clr

dir_path = os.path.dirname(os.path.realpath(__file__))
assemblydir = os.path.join(dir_path, 'PeaksimpleClient',
                           'PeaksimpleConnector.dll')

clr.AddReference(assemblydir)  # Add the assembly to python.NET
# Now that the Assembly has been added to python.NET,
# it can be imported like a normal module, though the name is different.
import Peaksimple  # Need add assembly before import. # noqa # type: ignore

# TODO some control file needs to be placed within package to work w/ any user.
# the default won't run from the repo for some reason
default_ctrl_file = ("C:\\Peak489Win10\\CONTROL_FILE\\"
                     "HayN_C2H2_Hydrogenation\\C2H2_Hydro_HayN_TCD_off.CON")


class GC_Connector():
    """
    Interface with SRI GC through Peaksimple interface via .dll file.

    If the assembly can't be found, find the .dll file, right click,
    properties, check "unblock", click apply

    Parameters
    ----------
    ctrl_file : str
        Full path to the initial ctrl file to use.
    """

    def __init__(self, ctrl_file=default_ctrl_file):
        """Connect to GC and load control file."""
        print('Connecting to Peaksimple...')
        # Define Attributes:
        # ------------------
        self.peaksimple = Peaksimple.PeaksimpleConnector()
        """Create instance of PeaksimpleConnector.
        :ref:`sri_gc_doc` for more info on methods."""

        self.ctrl_file = ctrl_file
        """str: Full path to control file to use."""

        self.sample_set_size = 4
        """
        int:  Number of GC collection for each time peaksimple.setRunning()
        is called. Can be changed per experiment by editing
        :attr:`Experiment.sample_set_size
        <catalight.equipment.experiment_control.Experiment.sample_set_size>`
        """

        # Init Procedure:
        # ---------------
        self.connect()
        self._sample_rate = 0  # Dummy value, reset when ctrl file loaded
        # Sends ctrl file to GC, updates self w/ new data
        print('Loading', ctrl_file)
        self.load_ctrl_file()

    # Properties
    # ----------
    # Makes min_sample_rate read only
    min_sample_rate = property(lambda self: self._min_sample_rate)
    """float, read-only: (min) Minimum setpoint for sample_rate.
    Sum of Channel 1 Time and Channel 1 Posttime from GC control file.
    Value automatically updates when :meth:`read_ctrl_file` is called."""

    # Setting sample rate changes when connected to GC
    @property
    def sample_rate(self):
        """
        Time between setting collections in minutes.

        The GC will collect data every sample_rate minutes when sample_set_size
        is set >1. If the min_sample_rate is less than the entered value,
        prints a warning and resets sample_rate to min_sample_rate.
        """
        return self._sample_rate

    @sample_rate.setter
    def sample_rate(self, value):
        if value >= self._min_sample_rate:
            self._sample_rate = value
        else:
            print('Minimum Sample Rate = %5.2f' % self.min_sample_rate)
            print('Sample rate set to minimum')
            self._sample_rate = self.min_sample_rate

    # Methods:
    # --------
    def update_ctrl_file(self, data_file_path):
        """
        Write over the currently loaded ctrl file to update settings.

        Goes line by line through the control file defined by ctrl_file attr
        and rewrites the line to set appropriate options. Ensures postrun cycle
        & repeat, autoincrement, and save image, data, & results are all
        turned on. Updates data file path to that provided. Sets postrun repeat
        to sample_set_size.

        Parameters
        ----------
        data_file_path : str
            Full path to the directory in which you'd like to save data files.
        """
        with open(self.ctrl_file, 'r+') as ctrl_file:
            new_ctrl_file = []
            for line in ctrl_file:  # read values after '=' line by line
                if re.search('<DATA FILE PATH>=', line):
                    line = ('<DATA FILE PATH>=' + data_file_path + '\n')
                elif re.search('<CHANNEL 1 POSTRUN CYCLE>=', line):
                    line = ('<CHANNEL 1 POSTRUN CYCLE>=1\n')
                elif re.search('<CHANNEL 1 POSTRUN REPEAT>=', line):
                    line = ('<CHANNEL 1 POSTRUN REPEAT>='
                            + str(self.sample_set_size) + '\n')

                elif re.search('<CHANNEL 1 FILE>=', line):
                    line = ('<CHANNEL 1 FILE>=FID\n')
                elif re.search('<CHANNEL 1 POSTRUN SAVE DATA>=', line):
                    line = ('<CHANNEL 1 POSTRUN SAVE DATA>=1\n')
                elif re.search('<CHANNEL 1 POSTRUN SAVE RESULTS>=', line):
                    line = ('<CHANNEL 1 POSTRUN SAVE RESULTS>=1\n')
                elif re.search('<CHANNEL 1 POSTRUN AUTOINCREMENT>=', line):
                    line = ('<CHANNEL 1 POSTRUN AUTOINCREMENT>=1\n')
                elif re.search('<CHANNEL 1 POSTRUN SAVE IMAGE>=', line):
                    line = ('<CHANNEL 1 POSTRUN SAVE IMAGE>=1\n')

                elif re.search('<CHANNEL 2 FILE>=', line):
                    line = ('<CHANNEL 2 FILE>=TCD\n')
                elif re.search('<CHANNEL 2 POSTRUN SAVE DATA>=', line):
                    line = ('<CHANNEL 2 POSTRUN SAVE DATA>=1\n')
                elif re.search('<CHANNEL 2 POSTRUN SAVE RESULTS>=', line):
                    line = ('<CHANNEL 2 POSTRUN SAVE RESULTS>=1\n')
                elif re.search('<CHANNEL 2 POSTRUN AUTOINCREMENT>=', line):
                    line = ('<CHANNEL 2 POSTRUN AUTOINCREMENT>=1\n')
                elif re.search('<CHANNEL 2 POSTRUN SAVE IMAGE>=', line):
                    line = ('<CHANNEL 2 POSTRUN SAVE IMAGE>=1\n')

                new_ctrl_file += line
            ctrl_file.seek(0)
            ctrl_file.writelines(new_ctrl_file)

        print(self.ctrl_file)
        self.load_ctrl_file()

    def load_ctrl_file(self, max_tries=3):
        """
        Load a new ctrl file to the GC based on .ctrl_file attr.

        Attempt to load a new control file. Waits 5 seconds after loading
        then calls read_ctr_file()

        Parameters
        ----------
        max_tries : int, optional
            Number of attempts to make before aborting. The default is 3.

        Raises
        ------
        Peaksimple.ConnectionWriteFailedException:
            Somewhat randomly occurring error usually fixed by reattempting.
        Peaksimple.NoConnectionException:
            Communication to peaksimple lost. Likely need to reboot peaksimple.
        """
        print('Loading Control File...')
        for attempt in range(0, max_tries):
            try:
                self.peaksimple.LoadControlFile(self.ctrl_file)
                print('Successful!')
                break
            except Peaksimple.ConnectionWriteFailedException:
                print('Write error. Retrying...')
                time.sleep(1)
                continue
            except Peaksimple.NoConnectionException:
                print('Write Error: GC Not Connected')
                break
        time.sleep(5)  # I think peaksimple is cranky when rushed
        self.read_ctrl_file()

    def read_ctrl_file(self):
        """
        Read loaded control file and updates object with values from file.

        Currently updates min_sample_rate property by pulling run time postrun
        cycle time from control file. Updates sample_rate if it is faster than
        the new min_sample_rate

        Returns
        -------
        None.

        """
        with open(self.ctrl_file, 'r+') as ctrl_file:
            for line in ctrl_file:  # read values after '=' line by line

                if re.search('<CHANNEL 1 TIME>=', line):
                    run_time = int(line.split('=')[-1].strip(' \n'))

                elif re.search('<CHANNEL 1 POSTRUN CYCLE TIME>=', line):
                    post_time = int(line.split('=')[-1].strip(' \n'))

            self._min_sample_rate = (run_time + post_time) / 1000

            if self.sample_rate < self.min_sample_rate:
                self.sample_rate = self.min_sample_rate

    def is_running(self, max_tries=3):
        """
        Request status of GC collection (running or not).

        Parameters
        ----------
        max_tries : int, optional
            Number of attempts to make before aborting. The default is 3.

        Raises
        ------
        Exception
            Exception raised on calling IsRunning() are suppressed up until
            max_tries is reached.

        Returns
        -------
        result : bool
            When successful, returns the status of GC collection (on/off)

        """
        for attempt in range(1, max_tries + 1):
            try:
                result = self.peaksimple.IsRunning(1)
                return result
                break
            except Exception as e:
                print(e)
                if attempt < max_tries:
                    print('Retrying...')
                    time.sleep(1)
                    continue
                else:
                    print('Cannot Resolve')
                    raise e

    def connect(self, max_tries=1):
        """
        Attempt to connect to running peaksimple software.

        Parameters
        ----------
        max_tries : int, optional
            Number of attempts to make before aborting. The default is 1.

        Raises
        ------
        Peaksimple.ConnectionFailedException
            Connection to GC was unsuccessful.

        Returns
        -------
        None.

        """
        for attempt in range(1, max_tries + 1):
            try:
                self.peaksimple.Connect()
                print('Connected!')
                break
            except Peaksimple.ConnectionFailedException:
                if attempt < max_tries:
                    print('Connection error. Retrying...')
                    time.sleep(1)
                    continue
                else:
                    print('Cannot Connect :(')
                    raise Peaksimple.ConnectionFailedException

    def disconnect(self, max_tries=1):
        """
        Disconnect from peaksimple software. Reconnection still usually fails.

        Runs Peaksimple.PeaksimpleConnector.Disconnect(), but attempts to
        reconnect still fail. The peaksimple software needs to be manually
        closed to reconnect. Note, programmatic attempts to close peaksimple
        using subprocesses have also failed.

        Parameters
        ----------
        max_tries : int, optional
            Number of attempts to make before aborting. The default is 1.

        Returns
        -------
        None.

        """
        try:
            self.peaksimple.Disconnect()
            print('Disconnected!')
            # TODO would deleting peaksimple here fix reconnect issues?
        except Exception as e:
            print(e)


if __name__ == "__main__":
    print(dir_path)
    gc1 = GC_Connector()
    data_path = 'C:\\Peak489Win10\\GCDATA\\20221117_CodeTest'
    gc1.sample_set_size = 2
    gc1.update_ctrl_file(data_path)
    gc1.peaksimple.SetRunning(1, True)
    time.sleep(10)
    for n in range(0, 31):
        print(time.ctime(), '---', gc1.peaksimple.IsRunning(1))
        # IsRunning returns false inbetween runs. Returns True during run
        time.sleep(60)
    print('Finished')



import glob
import os
import sys
from setuptools import setup
from configparser import ConfigParser

# Get some values from the setup.cfg
conf = ConfigParser()
conf.read(['setup.cfg'])
options = dict(conf.items('options'))

install_requires = [s.strip() for s in options.get('install_requires').split('\n')]

# Check if pyqt5 is already installed, whether by pip or some other manager.
# If so, avoid trying to install it again.
try:
    import PyQt5  # noqa
    hasPyQt = True
except ImportError:
    hasPyQt = False
    pass
is_conda = os.path.exists(os.path.join(sys.prefix, 'conda-meta'))
print('is_conda, hasPyQt')
print(is_conda)
print(hasPyQt)
# pyqt5 is named pyqt5 on pypi and pyqt on conda
# due to possible conflicts, skip the pyqt5 requirement in conda environments
# that already have pyqt
# See: https://github.com/biolab/orange3/pull/5593/commits/816f623dd8ffc530e617928eeac67821cd6c3075
if not (hasPyQt and is_conda):
    install_requires.append('PyQt5')
else:
    print('Conda version of PyQt5 already installed')
    if 'PyQt5' in install_requires:
        install_requires.remove('PyQt5')

setup(install_requires=install_requires)



"""
Executable script to run basic analysis on existing data sets.

Created on Fri Jan  6 12:11:00 2023.
@author: Briley Bourgeois
"""
import sys

import matplotlib.pyplot as plt
import pandas as pd
from PyQt5.QtWidgets import QFileDialog, QApplication, QDialog
import catalight.analysis as analysis
from catalight.analysis.user_inputs import (DirectorySelector,
                                            PlotOptionsDialog,
                                            PlotOptionList)


def get_user_inputs(starting_dir=None, cal_folder=None):
    """
    Take user input using GUI for running initial analysis.

    Parameters
    ----------
    starting_dir : str, optional
        Path to initialize file dialog in. The default is None.
    cal_folder : str, optional
        Path containing calibration files. The default in None.

    Returns
    -------
    expt_dirs : list of str
        Paths to main directories for running analysis.
    calDF : pandas.DataFrame
        Formatted DataFrame containing gc calibration data.
        Specific to control file used!
        Format [ChemID, slope, intercept, start, end]
    response_dict : dict
        Dictionary of user plot options.
        Dict Keys depend on options proved to PlotOptionsDialog.
        {'overwrite': bool, 'basecorrect': bool, 'reactant': str,
        'target_molecule': str, 'mole_bal': str, 'figsize': tuple,
        'savedata': bool, 'switch_to_hours': float}

    """
    app = QApplication(sys.argv)  # noqa
    # Prompt user to select calibration file
    cal_file = QFileDialog.getOpenFileName(None, 'Select Calibration File',
                                           cal_folder, "csv file (*.csv)")[0]
    calDF = pd.read_csv(cal_file, delimiter=',', index_col='Chem ID')

    # Prompt user to select multiple experiment directories
    data_dialog = DirectorySelector(starting_dir)
    if data_dialog.exec_() == QDialog.Accepted:
        expt_dirs = data_dialog.get_output()

    # Edit Options specifically for initial analysis dialog
    include_dict = {'overwrite': True, 'basecorrect': True, 'reactant': True,
                    'target_molecule': True, 'mole_bal': True, 'figsize': True,
                    'savedata': True, 'switch_to_hours': True}
    options = PlotOptionList()  # Create default gui options list
    options.change_includes(include_dict)  # Modify gui components
    options_dialog = PlotOptionsDialog(options)  # Build dialog w/ options
    if options_dialog.exec_() == PlotOptionsDialog.Accepted:
        response_dict = options.values_todict()

    return expt_dirs, calDF, response_dict


def main(main_dirs, calDF, reactant, target_molecule, mole_bal='c',
         figsize=(6.5, 4.5), savedata=False, switch_to_hours=2,
         overwrite=False, basecorrect=True):
    """
    Run initial analysis.

    Finds existing expt_log.txt files within the provided directories, creates
    expt objects, runs analysis.run_analysis and analysis.plot_expt_summary
    where appropriate based on input parameters.

    Parameters
    ----------
    main_dirs : list of str
        Paths to main directories for running analysis. All expt_log.txt files
        within the main directory will be analyzed.
    calDF : pandas.DataFrame
        Formatted DataFrame containing gc calibration data.
        Specific to control file used!
        Format [ChemID, slope, intercept, start, end]
    new_x : list
        Array of new data to be used inplace of old x axis
    units : str
        new units for new x axis data
    reactant : str
        String identity of reactant molecule to track. Must match what exists
        in the calibration file exactly.
    target_molecule : str
        String identity of the target to use when calculating selectivity. Must
        match what exists in the calibration file exactly.
    mole_bal : str, optional
        Code will perform a mole balance for the element provided.
        The default is 'c'. (i.e. carbon balance)
    figsize : tuple, optional
         Desired size of output figure in inches (x,y), Default is (6.5, 4.5).
         figsize suggestions:
         1/2 slide = (6.5, 4.5);  1/6 slide = (4.35, 3.25);
         1/4 slide =  (5, 3.65); Full slide =    (9, 6.65);
    savedata : bool, optional
        Indicates whether or not to save data. The default is 'True'.
    switch_to_hours : float, optional
        Time in hours when the output should switch units to
        hours instead of minutes. The default is 2.
    overwrite : bool, optional
        True will rerun calculations for experiments containing avg_conc.csv
        files. False will only plot these. The default is False.
    basecorrect : bool, optional
        True will perform baseline correction on GC data. The default is True.

    Returns
    -------
    None.

    """
    plt.ioff()  # suppress plot windows
    options = (reactant, target_molecule, mole_bal,
               figsize, savedata, switch_to_hours)
    filepaths = analysis.tools.list_matching_files(main_dirs,
                                                   'expt_log', '.txt')
    expts = analysis.tools.list_expt_obj(filepaths)
    print(expts)
    for expt in expts:
        has_data = analysis.tools.list_matching_files(expt.results_path,
                                                      'avg_conc', '.csv')
        if not has_data or overwrite:  # skip if has data and overwrite=False
            calculations = analysis.tools.run_analysis(expt, calDF,
                                                       basecorrect, savedata)
            (concentrations, avg, std) = calculations

        (ax1, ax2, ax3) = analysis.plotting.plot_expt_summary(expt, calDF,
                                                              *options)
    print('Finished Initial Analysis')
    # These outputs are defined to make it a little easier for output
    # if that is ever desired. Otherwise open pickled figs!!


if __name__ == "__main__":

    plt.close('all')
    expt_dirs, calDF, response_dict = get_user_inputs()
    main(expt_dirs, calDF, **response_dict)



"""
Graphical User Interface Module.

Imports a GUI design created using QT Designer. Connects signals/slots and
adds additional functionality.

Created on Sun Feb 20 18:45:10 2022.

@author: Briley Bourgeois
"""
import os
import subprocess
import sys
import time

import matplotlib
import numpy as np
import pandas as pd
import psutil
from alicat import FlowController
from catalight.equipment.alicat_MFC.gas_control import Gas_System
from catalight.equipment.diode_laser.diode_control import Diode_Laser
from catalight.equipment.harrick_watlow.heater_control import Heater
from catalight.equipment.sri_gc.gc_control import GC_Connector
from catalight.equipment.experiment_control import Experiment

import matplotlib.pyplot as plt
from matplotlib.backends.backend_qt5agg import \
    NavigationToolbar2QT as NavigationToolbar
from PyQt5.QtCore import (QObject, QRunnable, Qt, QThreadPool, QTimer,
                          pyqtSignal, pyqtSlot)
from PyQt5.QtGui import QTextCursor
from PyQt5.QtWidgets import (QAbstractItemView, QApplication, QComboBox,
                             QDialogButtonBox, QDoubleSpinBox, QFileDialog,
                             QLabel, QListWidgetItem, QMainWindow, QMessageBox,
                             QPushButton, QSpinBox)
from PyQt5.uic import loadUi
matplotlib.use('Qt5Agg')


class MainWindow(QMainWindow):
    """Subclass QMainWindow to customize your application's main window."""
    progress_signal = pyqtSignal(int)  #: triggered to update progress bar
    change_color_signal = pyqtSignal(object, str)  #: update QLabel font color

    def __init__(self):
        super().__init__()
        loadUi(r'.\gui_components\\reactorUI.ui', self)

        # Initilize GUI
        try:
            peaksimple = self.open_peaksimple(r"C:\Peak489Win10\Peak489Win10.exe")
        except FileNotFoundError:
            print('Peaksimple.exe not found')

        # sys.stdout = EmittingStream(self.consoleOutput)
        self.timer = QTimer(self)
        self.threadpool = QThreadPool()
        # Pass the function to execute
        # Any other args, kwargs are passed to the run function
        self.run_study_thread = Worker(self.start_study)
        self.manual_ctrl_thread = Worker(self.manual_ctrl_eqpt)
        self.run_study_thread.setAutoDelete(False)
        self.manual_ctrl_thread.setAutoDelete(False)

        # Initilize equipment
        self.init_equipment()
        self.init_study_tab()
        self.init_design_tab()
        self.init_manual_ctrl_tab()
        self.connect_manual_ctrl()
        self.init_figs()
        self.set_form_limits()

        self.file_browser = QFileDialog()
        self.emergencyStop.clicked.connect(self.emergency_stop)

    def normalOutputWritten(self, text):
        """Append console ouput to the QTextEdit."""
        # Maybe QTextEdit.append() works as well, but this is how I do it:
        cursor = self.consoleOutput.textCursor()
        cursor.movePosition(QTextCursor.End)
        cursor.insertText(text)
        self.consoleOutput.setTextCursor(cursor)
        self.consoleOutput.ensureCursorVisible()

    # Button Definitions:
    # -------------------
    def add_expt(self):
        """Create a new experiment object and add it to GUI."""
        item = QListWidgetItem('Undefined Experiment', self.listWidget)
        expt = Experiment()
        item.setData(Qt.UserRole, expt)

    def delete_expt(self):
        """Delete currently selected item from listWidget."""
        item = self.listWidget.currentItem()
        self.listWidget.takeItem(self.listWidget.row(item))

    def start_study(self):
        """
        Loop through experiment objects, update, and call .run_experiment().

        Takes experiment objects in listWidget, assigns eqpt, generates
        experiment name and save path, then call expt.run_experiment().
        Shuts down at end.
        """
        self.toggle_controls(True)
        expt_list = [self.listWidget.item(x).data(Qt.UserRole)
                     for x in range(self.listWidget.count())]
        eqpt_list = [self.gc_connector, self.laser_controller,
                     self.gas_controller, self.heater]
        sample_name = (self.sample_name.text()
                       + str(self.sample_mass.value()))
        main_fol = os.path.join(r'C:\Peak489Win10\GCDATA', sample_name)
        os.makedirs(main_fol, exist_ok=True)

        for expt in expt_list:

            expt.sample_name = sample_name
            expt.create_dirs(main_fol)
            expt.update_eqpt_list(eqpt_list)
            print(expt.expt_name)
            print(expt.sample_name)
            expt.run_experiment()
        self.shut_down()
        self.toggle_controls(False)

    def select_ctrl_file(self):
        """Prompt user to selector GC control file if gc is connected."""
        if self.gc_Status.isChecked():
            options = self.file_browser.Options()
            options |= self.file_browser.DontUseNativeDialog
            filePath = self.file_browser \
                           .getOpenFileName(self, 'Select Control File',
                                            "C:\\Peak489Win10\\CONTROL_FILE",
                                            "Control files (*.CON)")[0]
            print(filePath)
            self.ctrl_path.setText(filePath)
            self.gc_connector.ctrl_file = filePath
            load_ctrl_file_thread = Worker(self.gc_connector.load_ctrl_file)
            self.threadpool.start(load_ctrl_file_thread)
            load_ctrl_file_thread.signal.finished.connect(self.set_form_limits)

        else:
            print('GC Not Connected')

    def select_cal_file(self):
        """
        Prompt user to select calibration file.

        If gas system is connected, runs "set_calibration()" method on each mfc
        Adds 'CalGas' option to mfc gas type drop downs
        """
        options = self.file_browser.Options()
        options |= self.file_browser.DontUseNativeDialog
        filePath = self.file_browser \
                       .getOpenFileName(self, 'Select Calibration File',
                                        "C:\\Peak489Win10\\CONTROL_FILE",
                                        "Control files (*.csv)")[0]
        print(filePath)
        self.cal_path.setText(filePath)
        calDF = pd.read_csv(filePath, delimiter=',', index_col='Chem ID')
        if self.gas_Status.isChecked():
            attribute_list = vars(self.gas_controller)
            for key in attribute_list:
                attr = attribute_list[key]
                if isinstance(attr, FlowController):
                    self.gas_controller.set_calibration_gas(attr, calDF,
                                                             fill_gas='Ar')

            self.setGasAType.insertItem(0, 'CalGas')
            self.setGasBType.insertItem(0, 'CalGas')
            self.setGasCType.insertItem(0, 'CalGas')
            self.setGasDType.insertItem(0, 'CalGas')
            self.manualGasAType.insertItem(0, 'CalGas')
            self.manualGasBType.insertItem(0, 'CalGas')
            self.manualGasCType.insertItem(0, 'CalGas')
            self.manualGasDType.insertItem(0, 'CalGas')

    def reset_eqpt(self):
        """Disconnects from equipment and attempts to reconnect."""
        print('Resetting Equipment Connections')
        self.disconnect()
        self.init_equipment()
        self.init_manual_ctrl_tab()
        self.set_form_limits()

    # Initializing Tabs:
    # ------------------
    def init_equipment(self):
        """
        Try to connect to each piece of equipment.

        Adjusts (object)_Status.setChecked if successful.
        If Exception is thrown, setChecked(0) and print Exception
        Calls set_form_limits() at end
        """
        # Initialize Equipment
        # Try to connect to each device, mark indicator off on failed connect
        try:
            self.gc_connector = GC_Connector()
            self.gc_Status.setChecked(1)
        except Exception as e:
            print(e)
            self.gc_Status.setChecked(0)
        try:
            self.gas_controller = Gas_System()
            self.gas_Status.setChecked(1)
        except Exception as e:
            print(e)
            self.gas_Status.setChecked(0)
        try:
            self.heater = Heater()
            self.heater_Status.setChecked(1)
        except Exception as e:
            print(e)
            self.heater_Status.setChecked(0)
        try:
            self.laser_controller = Diode_Laser()
            # Check if output is supported by DAQ
            if self.laser_controller._ao_info.is_supported:
                self.diode_Status.setChecked(1)
            else:
                self.diode_Status.setChecked(0)
        except Exception as e:
            print(e)
            self.diode_Status.setChecked(0)

        self.set_form_limits()

    def init_study_tab(self):
        """Connect Study Overview Tab Contents (signals/slots)."""
        self.setWindowTitle("CataLight")
        self.butAddExpt.clicked.connect(self.add_expt)
        self.butDelete.clicked.connect(self.delete_expt)
        self.butStart.clicked.connect(lambda: self.threadpool.start(self.run_study_thread))
        self.listWidget.itemClicked.connect(self.display_expt)
        self.listWidget.setDragDropMode(QAbstractItemView.InternalMove)
        self.findCalFile.clicked.connect(self.select_cal_file)
        self.findCtrlFile.clicked.connect(self.select_ctrl_file)

    def init_design_tab(self):
        """
        Connect Experiment Design Tab Contents (signals/slots).

        On first call, populates expt_type drop down on GUI.
        Insert possible gasses to combo boxes.
        Initializes widgets used for defining comp_sweep experiments.
        """
        self.expt_types.currentIndexChanged.connect(self.update_expt)
        # On first run, this should populate expt drop down on GUI
        if self.expt_types.count() < len(Experiment().expt_list['Expt Name']):
            self.expt_types.addItem('Undefined')
            expt_list = Experiment().expt_list['Expt Name'].to_list()
            self.expt_types.addItems(expt_list)
        self.setTemp.valueChanged.connect(self.update_expt)
        self.setPower.valueChanged.connect(self.update_expt)
        self.setFlow.valueChanged.connect(self.update_expt)
        # TODO This needs to have a min set by ctrl file
        self.setSampleRate.valueChanged.connect(self.update_expt)
        self.setSampleSize.valueChanged.connect(self.update_expt)
        # TODO get from heater?
        self.setRampRate.valueChanged.connect(self.update_expt)
        self.setTSteady.valueChanged.connect(self.update_expt)
        self.setBuffer.valueChanged.connect(self.update_expt)
        self.setGasAComp.valueChanged.connect(self.update_expt)
        self.setGasAType.currentIndexChanged.connect(self.update_expt)
        self.setGasBComp.valueChanged.connect(self.update_expt)
        self.setGasBType.currentIndexChanged.connect(self.update_expt)
        self.setGasCComp.valueChanged.connect(self.update_expt)
        self.setGasCType.currentIndexChanged.connect(self.update_expt)
        self.setGasDComp.valueChanged.connect(self.update_expt)
        self.setGasDType.currentIndexChanged.connect(self.update_expt)
        # lineEdit.editingFinished() good one
        self.IndVar_start.valueChanged.connect(self.update_expt)
        self.IndVar_stop.valueChanged.connect(self.update_expt)
        self.IndVar_step.valueChanged.connect(self.update_expt)
        self.setGasAType.insertItems(0, Gas_System.factory_gasses)
        self.setGasBType.insertItems(0, Gas_System.factory_gasses)
        self.setGasCType.insertItems(0, Gas_System.factory_gasses)
        self.setGasDType.insertItems(0, Gas_System.factory_gasses)
        if self.gas_Status.isChecked():  # sets gas type to last used
            flow_dict = self.gas_controller.read_flows()
            self.setGasAType.setCurrentText(flow_dict['mfc_A']['gas'])
            self.setGasBType.setCurrentText(flow_dict['mfc_B']['gas'])
            self.setGasCType.setCurrentText(flow_dict['mfc_C']['gas'])
            self.setGasDType.setCurrentText(flow_dict['mfc_D']['gas'])

        # Label independent variable widgets as default grid
        widgets = [[self.label_78, self.label_79, self.label_80],
                   [self.IndVar_start, self.IndVar_stop, self.IndVar_step]]
        self.default_grid_widgets = widgets

        # create initial list of buttons to be added into grid layout when
        # comp sweep is selected
        self.comp_sweep_widgets = [[QLabel('-'), QLabel('Multiplier'),
                                    QLabel('Status'), QLabel('Start'),
                                    QLabel('Stop'), QLabel('Step')]]

        combobox_options = ['-', 'Fixed', 'Fill', 'Ind. Variable', 'Multiple']
        for i in range(1, 5):
            self.comp_sweep_widgets.append([QLabel(('Gas %i' % i)),
                                            QDoubleSpinBox(),
                                            QComboBox(),
                                            QDoubleSpinBox(),
                                            QDoubleSpinBox(),
                                            QDoubleSpinBox()])

            # Add drop down items to combo box that was just added to list
            self.comp_sweep_widgets[i][2].addItems(combobox_options)
            self.comp_sweep_widgets[i][2].currentIndexChanged.connect(self.update_expt)
            for j in [1, 3, 4, 5]:
                self.comp_sweep_widgets[i][j].valueChanged.connect(self.update_expt)
                self.comp_sweep_widgets[i][j].setMaximum(100)

    def init_manual_ctrl_tab(self):
        """
        Initialize Manual Control Tab.

        Blocks tabWidget updates, puts intial values in manual tab widgets.
        Insert possible gasses to combo boxes. The update process for live
        readings for both the manual and live view tabs is created/connected
        in the connect_manual_ctrl() method.
        """
        self.tabWidget.setUpdatesEnabled(False)  # Block Signals during update
        # Initialize Values for gas controller
        if self.gas_Status.isChecked():
            flow_dict = self.gas_controller.read_flows()
            tot_flow = (flow_dict['mfc_A']['setpoint']
                        + flow_dict['mfc_B']['setpoint']
                        + flow_dict['mfc_C']['setpoint']
                        + flow_dict['mfc_D']['setpoint'])
            if tot_flow == 0:
                self.manualGasAComp.setValue(0)
                self.manualGasBComp.setValue(0)
                self.manualGasCComp.setValue(0)
                self.manualGasDComp.setValue(0)
            else:
                self.manualGasAComp.setValue(flow_dict['mfc_A']['setpoint'] / tot_flow * 100)
                self.manualGasBComp.setValue(flow_dict['mfc_B']['setpoint'] / tot_flow * 100)
                self.manualGasCComp.setValue(flow_dict['mfc_C']['setpoint'] / tot_flow * 100)
                self.manualGasDComp.setValue(flow_dict['mfc_D']['setpoint'] / tot_flow * 100)

            self.manualGasAType.insertItems(0, Gas_System.factory_gasses)
            self.manualGasBType.insertItems(0, Gas_System.factory_gasses)
            self.manualGasCType.insertItems(0, Gas_System.factory_gasses)
            self.manualGasDType.insertItems(0, Gas_System.factory_gasses)
            self.manualGasAType.setCurrentText(flow_dict['mfc_A']['gas'])
            self.manualGasBType.setCurrentText(flow_dict['mfc_B']['gas'])
            self.manualGasCType.setCurrentText(flow_dict['mfc_C']['gas'])
            self.manualGasDType.setCurrentText(flow_dict['mfc_D']['gas'])
            self.manualFlow.setValue(tot_flow)

        if self.heater_Status.isChecked():  # Initialize Values for Heater
            self.manualTemp.setValue(self.heater.read_setpoint())
            self.manualRamp.setValue(self.heater.ramp_rate)

        if self.diode_Status.isChecked():  # Initialize Values for Laser
            self.manualPower.setValue(self.laser_controller.get_output_power())

        self.tabWidget.setUpdatesEnabled(True)  # Allow signals again

    def sum_spinboxes(self, spinboxes, qlabel):
        """
        Take list of spinboxes, get values, write sum to qlabel.

        Parameters
        ----------
        spinboxes : list of PyQt5.QtWidgets.QSpinBoxes
            spinboxes to sum
        qlabel : PyQt5.QtWidgets.QLabel
            qlabel to write to

        Returns
        -------
        None.

        """
        values = self.values_from_spinboxes(spinboxes)
        qlabel.setText('%.2f' % sum(values))

    def values_from_spinboxes(self, spinboxes):
        """
        Give list of spinboxes, return list of values.

        Parameters
        ----------
        spinboxes : list
            list of spinboxes you'd like to get values from

        Returns
        -------
        gas_comp : list
            list of the values from the spinboxes supplied

        """
        values = []
        for entry in spinboxes:
            values.append(entry.value())
        return values

    def connect_manual_ctrl(self):
        """
        Connect Manual Control (signals/slots).

        Connect buttons within manual control tab to corresponding functions.
        Also initialize a timer that calls update_eqpt_status(), updating both
        the manua_ctrl tab and live view
        """
        # Connect buttons in manual ctrl tab
        self.buttonBox.button(QDialogButtonBox.Apply).clicked \
            .connect(lambda: self.threadpool.start(self.manual_ctrl_thread))

        self.buttonBox.button(QDialogButtonBox.Reset).clicked \
            .connect(self.init_manual_ctrl_tab)

        self.eqpt_ReconnectBut.clicked.connect(self.reset_eqpt)

        # Connect timer for live feed
        self.timer = QTimer(self)
        self.timer.start(500)  # timer connected to update in init_manual_ctrl
        self.timer.timeout.connect(self.update_eqpt_status)

        # Redefine function and arguments for brevity.
        gas_spinboxes = [self.manualGasAComp, self.manualGasBComp,
                         self.manualGasCComp, self.manualGasDComp]
        func = self.sum_spinboxes
        args = (gas_spinboxes, self.manualCompSum)

        # Connect with lambda functions so arguments can be passed.
        self.manualGasAComp.valueChanged.connect(lambda: func(*args))
        self.manualGasBComp.valueChanged.connect(lambda: func(*args))
        self.manualGasCComp.valueChanged.connect(lambda: func(*args))
        self.manualGasDComp.valueChanged.connect(lambda: func(*args))
        self.progress_signal.connect(self.progressBar.setValue)
        self.change_color_signal.connect(self.change_label_color)

    def init_figs(self):
        """
        Initialize figure canvas.

        Creates figure objects, adds to gui, adds navigation bar
        """
        # Create New figure to share
        self.figure = plt.figure()
        # Reset fig in canvases
        self.plotWidgetStudy.canvas.figure = self.figure
        self.plotWidgetDesign.canvas.figure = self.figure
        # Add navigation bar beneath each figure widget
        self.verticalLayoutStudyFig.addWidget(NavigationToolbar(self.plotWidgetStudy.canvas, self))
        self.verticalLayoutDesignFig.addWidget(NavigationToolbar(self.plotWidgetDesign.canvas, self))

    def set_form_limits(self):
        """
        Update the limits of widgets within the GUI.

        Use this function to set widget min/max values. These can either be
        pulled from eqpt objects to have equipment specific limits, or hard
        coded. Alternativly, many limits are set using QtDesigner
        Users are encourage to edit this for specific equipment setups.
        """
        if self.gc_Status.isChecked():
            self.setSampleRate.setMinimum(self.gc_connector.min_sample_rate)

        if self.heater_Status.isChecked():
            self.manualTemp.setMaximum(450)

        if self.gas_Status.isChecked():
            self.manualFlow.setMaximum(350)

    # Updating Tabs/Objects:
    # ----------------------
    def display_expt(self):
        """Update GUI display when new expt is selected in listWidget."""
        self.tabWidget.setUpdatesEnabled(False)
        global update_flag
        update_flag = False
        item = self.listWidget.currentItem()
        expt = item.data(Qt.UserRole)

        self.expt_types.setCurrentText(expt.expt_type)

        if expt.ind_var != 'Undefined':
            values = getattr(expt, expt.ind_var)
            has_data = len(values) > 1
            if has_data and (expt.expt_type in ['comp_sweep', 'calibration']):

                for i in range(1, len(self.comp_sweep_widgets)):
                    # display only version of composition values
                    self.comp_sweep_widgets[i][1].setValue(0)
                    self.comp_sweep_widgets[i][2].setCurrentText('-')
                    self.comp_sweep_widgets[i][3].setValue(min(values[i]))
                    self.comp_sweep_widgets[i][4].setValue(max(values[i]))
                    self.comp_sweep_widgets[i][5].setValue(values[i][1] - values[i][0])

            elif has_data:
                self.IndVar_start.setValue(np.min(values))
                self.IndVar_stop.setValue(np.max(values))
                self.IndVar_step.setValue(values[1] - values[0])
            self.update_plot(expt)

        else:
            self.IndVar_start.setValue(0)
            self.IndVar_stop.setValue(0)
            self.IndVar_step.setValue(0)
            self.update_plot()

        self.setTemp.setValue(expt.temp[0])
        self.setPower.setValue(expt.power[0])
        self.setFlow.setValue(expt.tot_flow[0])

        self.setSampleRate.setValue(expt.sample_rate)  # This needs to have a min set by ctrl file
        self.setSampleSize.setValue(expt.sample_set_size)
        self.setRampRate.setValue(expt.heat_rate)  # get from heater?
        self.setBuffer.setValue(expt.t_buffer)
        self.setTSteady.setValue(expt.t_steady_state)
        self.setGasAComp.setValue(expt.gas_comp[0][0])
        self.setGasAType.setCurrentText(expt.gas_type[0])
        self.setGasBComp.setValue(expt.gas_comp[0][1])
        self.setGasBType.setCurrentText(expt.gas_type[1])
        self.setGasCComp.setValue(expt.gas_comp[0][2])
        self.setGasCType.setCurrentText(expt.gas_type[2])
        self.setGasDComp.setValue(expt.gas_comp[0][3])
        self.setGasDType.setCurrentText(expt.gas_type[3])
        self.update_ind_var_grid()
        comp_total = sum(expt.gas_comp[0])  # Calculate gas comp total
        self.designCompSum.setText('%.2f' % comp_total)
        self.tabWidget.setUpdatesEnabled(True)
        update_flag = True

    def update_expt(self):
        """
        Update Experiment object with GUI values.

        Populates the attributes of currently selected experiment with the
        values displayed in GUI. If valid independent variable is set, updates
        plot. Clears plot for invalid experiment.
        """
        # grab the data associated with selected listWidget item
        item = self.listWidget.currentItem()
        global update_flag
        # If there is data in listWidget item and now in the middle of updating
        if (item is not None) & update_flag:
            expt = item.data(Qt.UserRole)  # pull listWidgetItem data out
            # set the attributes of expt object based on GUI entries
            if self.expt_types.currentText() == 'Undefined':
                return  # Skip update
            expt.expt_type = self.expt_types.currentText()
            expt.temp[0] = self.setTemp.value()
            expt.power[0] = self.setPower.value()
            expt.tot_flow[0] = self.setFlow.value()
            expt.sample_rate = self.setSampleRate.value()  # This needs to have a min set by ctrl file
            expt.sample_set_size = self.setSampleSize.value()
            expt.heat_rate = self.setRampRate.value()  # get from heater?
            expt.t_buffer = self.setBuffer.value()
            expt.t_steady_state = self.setTSteady.value()
            expt.gas_comp[0][0] = self.setGasAComp.value()
            expt.gas_type[0] = self.setGasAType.currentText()
            expt.gas_comp[0][1] = self.setGasBComp.value()
            expt.gas_type[1] = self.setGasBType.currentText()
            expt.gas_comp[0][2] = self.setGasCComp.value()
            expt.gas_type[2] = self.setGasCType.currentText()
            expt.gas_comp[0][3] = self.setGasDComp.value()
            expt.gas_type[3] = self.setGasDType.currentText()
            comp_total = sum(expt.gas_comp[0])  # Calculate gas comp total
            self.designCompSum.setText('%.2f' % comp_total)

            expt._update_expt_name()  # autoname experiment
            item.setText(expt.expt_type + expt.expt_name)  # add name to listWidget

            # pull units based on expt class definitions, label in GUI
            units = (expt.expt_list['Units']
                     [expt.expt_list['Active Status']].to_string(index=False))
            self.label_IndVar.setText(expt.ind_var + ' [' + units + ']')

            self.update_ind_var_grid()  # why is this here??
            self.update_ind_var(expt)  # Checks for logical values before plot
            # self.update_plot(expt)

    def update_ind_var(self, expt):
        """
        Update independent variable if applicable.

        Checks for valid independent variable and updates the experiment
        object accordingly. Returns if invalid variable is supplied

        Parameters
        ----------
            expt : catalight.equipment.Experiment
                Experiment object containing the expt_type to be analyzed
        """
        # only actually updates ind var if number check out
        if expt.expt_type in ['comp_sweep', 'calibration']:
            widget_grid = self.comp_sweep_widgets
            row_types = ['null']  # first row of grid is QLabel
            num_rows = len(widget_grid)
            comp_list = [[]] * (num_rows - 1)  # preallocate comp_list size

            for i in range(1, num_rows):  # sweep rows of grid
                # ['null', combobox1, combobox2, etc]
                row_types.append(widget_grid[i][2].currentText())

            # Cases to Reject!!
            if '-' in row_types:
                return
            if (row_types.count('Fill') != 1) or \
               (row_types.count('Ind. Variable') != 1):
                self.update_plot()
                return

            row_types = np.array(row_types)  # to allow np.where()

            # This should execute 0 or 1 times if TODO double updating blocked
            for i in np.where(row_types == 'Ind. Variable')[0].tolist():
                ind_var_row = int(np.where(row_types == 'Ind. Variable')[0])

                # This block determines if sensible values are in ind_var
                start = widget_grid[ind_var_row][3].value()
                stop = widget_grid[ind_var_row][4].value() + 1
                step = widget_grid[ind_var_row][5].value()
                if (stop > start) and (step > 0):
                    ind_var = np.arange(start, stop, step)
                else:
                    return(False)

                ind_var = ind_var / 100  # convert from % to frac
                comp_list[ind_var_row - 1] = ind_var.tolist()

                # fill_vals = np.ones(len(ind_var)) # Fill values = 1 (100%)
                fill_vals = 1 - ind_var

            for i in np.where(row_types == 'Multiple')[0].tolist():
                comp_vals = ind_var * widget_grid[i][1].value()
                comp_list[i - 1] = comp_vals.tolist()
                fill_vals = fill_vals - comp_vals

            for i in np.where(row_types == 'Fixed')[0].tolist():
                comp_vals = np.ones(len(ind_var)) * widget_grid[i][3].value()
                comp_vals = comp_vals / 100  # convert from % to frac
                comp_list[i - 1] = comp_vals.tolist()
                fill_vals = fill_vals - comp_vals

            for i in np.where(row_types == 'Fill')[0].tolist():
                comp_list[i - 1] = fill_vals.tolist()

            if (fill_vals < 0).any():
                self.update_plot()  # Clears plot
                return  # Cancels update

            comp_list = np.array(comp_list).T  # Transpose comp_list
            comp_list = comp_list.tolist()  # return to list of lists
            setattr(expt, expt.ind_var, comp_list)

        elif (self.IndVar_stop.value() > self.IndVar_start.value()) and \
             (self.IndVar_step.value() > 0) and \
             (expt.expt_type != 'stability_test'):
            setattr(expt, expt.ind_var,
                    list(np.arange(self.IndVar_start.value(),
                                   self.IndVar_stop.value() + 1,
                                   self.IndVar_step.value())))
        else:
            self.update_plot()  # clear plot
            return  # reject update
        self.update_plot(expt)  # if either if statement was true

    def update_ind_var_grid(self):
        """
        Update widgets in independent variable grid.

        Chooses the correct widgets to display based on the experiment type
        currently selected in the GUI.
        """
        def clear_grid():
            # Remove and hide currently present widgets
            for i in range(self.gridLayout_9.rowCount()):
                for j in range(self.gridLayout_9.columnCount()):
                    item = self.gridLayout_9.itemAtPosition(i, j)
                    if item is not None:
                        widget = item.widget()
                        self.gridLayout_9.removeWidget(widget)
                        widget.setHidden(True)

        item = self.listWidget.currentItem()
        global update_flag
        # If there is data in listWidget item and now in the middle of updating
        if (item is not None) & update_flag:
            expt = item.data(Qt.UserRole)  # pull listWidgetItem data out
        else:
            return

        # if comp sweep and grid isn't set up for it
        if (expt.expt_type in ['comp_sweep', 'calibration']) and \
           (self.comp_sweep_widgets[0][0].isHidden()):

            clear_grid()
            # Add comp sweep specific widgets to layout
            for i in range(len(self.comp_sweep_widgets)):  # sweep items to add
                for j in range(len(self.comp_sweep_widgets[0])):
                    self.gridLayout_9.addWidget(self.comp_sweep_widgets[i][j], i, j)
                    self.comp_sweep_widgets[i][j].setHidden(False)

            if (expt.expt_type == 'calibration') and \
               (not os.path.isfile(self.cal_path.text())):
                msg = QMessageBox()
                msg.setIcon(QMessageBox.Warning)
                msg.setText('Warning: You must enter appropriate calibration '
                            'data file path on overview tab to run calibration!!!')
                msg.setWindowTitle('Calibration Data Path Warning')
                msg.exec()

        # if not comp sweep but grid is still set up
        elif (expt.expt_type not in ['comp_sweep', 'calibration']) and \
             (self.default_grid_widgets[0][0].isHidden()):

            clear_grid()
            # Add default widgets to layout, sweep items to add
            for i in range(len(self.default_grid_widgets)):
                for j in range(len(self.default_grid_widgets[0])):
                    widget = self.default_grid_widgets[i][j]
                    self.gridLayout_9.addWidget(widget, i, j)
                    self.default_grid_widgets[i][j].setHidden(False)

        # update active elements if comp_sweep is set up
        if (expt.expt_type in ['comp_sweep', 'calibration']) and \
           (self.default_grid_widgets[0][0].isHidden()):

            # sweep rows of grid
            for i in range(1, len(self.comp_sweep_widgets)):

                if self.comp_sweep_widgets[i][2].currentText() == 'Fill':
                    selection_rule = [False, False, False, False]

                elif self.comp_sweep_widgets[i][2].currentText() == 'Fixed':
                    selection_rule = [False, True, False, False]

                elif self.comp_sweep_widgets[i][2].currentText() == 'Multiple':
                    selection_rule = [True, False, False, False]

                elif self.comp_sweep_widgets[i][2].currentText() == 'Ind. Variable':
                    selection_rule = [False, True, True, True]

                else:  # This is likely currentText == '-'
                    selection_rule = [False, False, False, False]
                    self.comp_sweep_widgets[i][1].setValue(0)
                    self.comp_sweep_widgets[i][3].setValue(0)
                    self.comp_sweep_widgets[i][4].setValue(0)
                    self.comp_sweep_widgets[i][5].setValue(0)

                # Disable widgets based on selection rules
                self.comp_sweep_widgets[i][1].setEnabled(selection_rule[0])
                self.comp_sweep_widgets[i][3].setEnabled(selection_rule[1])
                self.comp_sweep_widgets[i][4].setEnabled(selection_rule[2])
                self.comp_sweep_widgets[i][5].setEnabled(selection_rule[3])

        self.gridLayout_9.update()

    def update_plot(self, expt=None):
        """
        Update the plots in GUI when experiment is changed.

        Parameters
        ----------
        expt : catalight.equipment.experiment_control.Experiment, optional
            Updates plot with contents of experiment if supplied.
            Scrubs plot if expt=None. The default is None.
        """
        if not expt:
            self.figure.clear()
        elif expt.ind_var:
            expt.plot_sweep(self.figure)
            self.figure.tight_layout()

        else:
            self.figure.clear()

        self.plotWidgetStudy.canvas.draw()
        self.plotWidgetStudy.canvas.show()
        self.plotWidgetDesign.canvas.draw()
        self.plotWidgetDesign.canvas.show()

    def update_eqpt_status(self):
        """
        Update the equipment live view.

        This function updates the live view of the equipment in both the
        manual control (1) and the live view (2) tabs
        """
        if self.diode_Status.isChecked():
            self.current_power_1.setText('%.2f' % self.laser_controller.get_output_power())
            self.current_power_2.setText('%.2f' % self.laser_controller.get_output_power())
            self.current_power_setpoint1.setText('%.2f' % self.laser_controller.P_set)
            self.current_power_setpoint2.setText('%.2f' % self.laser_controller.P_set)

        if self.heater_Status.isChecked():
            self.current_temp_1.setText('%.2f' % self.heater.read_temp())
            self.current_temp_2.setText('%.2f' % self.heater.read_temp())
            self.current_temp_setpoint1.setText('%.2f' % self.heater.read_setpoint())
            self.current_temp_setpoint2.setText('%.2f' % self.heater.read_setpoint())

        if self.gas_Status.isChecked():
            flow_dict = self.gas_controller.read_flows()
            self.current_gasA_comp_1.setText('%.2f' % flow_dict['mfc_A']['mass_flow'])
            self.current_gasA_pressure_1.setText('%.2f' % flow_dict['mfc_A']['pressure'])
            self.current_gasA_type_1.setText(flow_dict['mfc_A']['gas'])
            self.current_gasB_comp_1.setText('%.2f' % flow_dict['mfc_B']['mass_flow'])
            self.current_gasB_pressure_1.setText('%.2f' % flow_dict['mfc_B']['pressure'])
            self.current_gasB_type_1.setText(flow_dict['mfc_B']['gas'])
            self.current_gasC_comp_1.setText('%.2f' % flow_dict['mfc_C']['mass_flow'])
            self.current_gasC_pressure_1.setText('%.2f' % flow_dict['mfc_C']['pressure'])
            self.current_gasC_type_1.setText(flow_dict['mfc_C']['gas'])
            self.current_gasD_comp_1.setText('%.2f' % flow_dict['mfc_D']['mass_flow'])
            self.current_gasD_pressure_1.setText('%.2f' % flow_dict['mfc_D']['pressure'])
            self.current_gasD_type_1.setText(flow_dict['mfc_D']['gas'])
            self.current_gasE_flow_1.setText('%.2f' % flow_dict['mfc_E']['mass_flow'])
            self.current_gasE_pressure_1.setText('%.2f' % flow_dict['mfc_E']['pressure'])

            self.current_gasA_comp_2.setText('%.2f' % flow_dict['mfc_A']['mass_flow'])
            self.current_gasA_pressure_2.setText('%.2f' % flow_dict['mfc_A']['pressure'])
            self.current_gasA_type_2.setText(flow_dict['mfc_A']['gas'])
            self.current_gasB_comp_2.setText('%.2f' % flow_dict['mfc_B']['mass_flow'])
            self.current_gasB_pressure_2.setText('%.2f' % flow_dict['mfc_B']['pressure'])
            self.current_gasB_type_2.setText(flow_dict['mfc_B']['gas'])
            self.current_gasC_comp_2.setText('%.2f' % flow_dict['mfc_C']['mass_flow'])
            self.current_gasC_pressure_2.setText('%.2f' % flow_dict['mfc_C']['pressure'])
            self.current_gasC_type_2.setText(flow_dict['mfc_C']['gas'])
            self.current_gasD_comp_2.setText('%.2f' % flow_dict['mfc_D']['mass_flow'])
            self.current_gasD_pressure_2.setText('%.2f' % flow_dict['mfc_D']['pressure'])
            self.current_gasD_type_2.setText(flow_dict['mfc_D']['gas'])
            self.current_gasE_flow_2.setText('%.2f' % flow_dict['mfc_E']['mass_flow'])
            self.current_gasE_pressure_2.setText('%.2f' % flow_dict['mfc_E']['pressure'])

    def manual_ctrl_eqpt(self):
        """
        Update setpoint of equipment.

        Updates the setpoint of all equipment based on the current manual
        control values entered in the GUI. This method is usually called from a
        worker thread as it takes a long time to complete and uses many
        MainWindow attributes. As such, this method emits several signals to
        trigger visual updates in the GUI.

        .. Note::
            The visual updates need to be sent out as signals to not cause
            issues with updating the GUI outside the main thread!!

        """
        self.toggle_controls(True)
        comp_list = [self.manualGasAComp.value(),
                     self.manualGasBComp.value(),
                     self.manualGasCComp.value(),
                     self.manualGasDComp.value()]
        gas_list = [self.manualGasAType.currentText(),
                    self.manualGasBType.currentText(),
                    self.manualGasCType.currentText(),
                    self.manualGasDType.currentText()]
        tot_flow = self.manualFlow.value()

        if self.gas_Status.isChecked():
            self.gas_controller.set_gasses(gas_list)
            self.progress_signal.emit(25)
            self.gas_controller.set_flows(comp_list, tot_flow)
        self.progress_signal.emit(50)

        if self.diode_Status.isChecked():
            self.change_color_signal.emit(self.current_power_setpoint1, 'red')
            self.change_color_signal.emit(self.current_power_setpoint2, 'red')
            self.laser_controller.set_power(self.manualPower.value())
            self.change_color_signal.emit(self.current_power_setpoint1, 'white')  # noqa
            self.change_color_signal.emit(self.current_power_setpoint2, 'white')  # noqa
        self.progress_signal.emit(75)

        if self.heater_Status.isChecked():
            self.change_color_signal.emit(self.current_temp_setpoint1, 'red')
            self.change_color_signal.emit(self.current_temp_setpoint2, 'red')
            self.heater.ramp_rate = self.manualRamp.value()
            self.heater.ramp(self.manualTemp.value())
            self.change_color_signal.emit(self.current_temp_setpoint1, 'white')
            self.change_color_signal.emit(self.current_temp_setpoint2, 'white')

        self.progress_signal.emit(100)
        self.toggle_controls(False)

    def toggle_controls(self, value):
        """
        Toggle status of widgets in tabWidget.

        Parameters
        ----------
        value : bool
            True disables all widgets in tabWidget (not emergency stop)
            False enables all
        """
        group = [*self.tabWidget.findChildren(QDialogButtonBox),
                 *self.tabWidget.findChildren(QDoubleSpinBox),
                 *self.tabWidget.findChildren(QPushButton),
                 *self.tabWidget.findChildren(QComboBox),
                 *self.tabWidget.findChildren(QSpinBox)]

        for item in group:
            item.setDisabled(value)

    def change_label_color(self, label, new_color):
        label.setStyleSheet('Color: ' + new_color)

    def open_peaksimple(self, path_name):
        """
        Use subprocess package to open peaksimple instance.

        Checks for running version of peaksimple first. Prints warning and
        returns if peaksimple is running. Using process.kill() doesn't give
        expected results. Peaksimple API gives error when trying to reconnect
        if you don't close the window manually. Still searching for solution.

        Parameters
        ----------
        path_name : str
            full path to PeakSimple executable

        Returns
        -------
        process : subprocess.process
            returns a process object for peaksimple instance
        """
        for process in psutil.process_iter():
            if 'Peak489Win10' in process.name():
                # process.kill()
                print('please close peaksimple and reconnect')
                time.sleep(5)
                return

        process = subprocess.Popen(path_name)
        time.sleep(5)
        return process

    def closeEvent(self, *args, **kwargs):
        """Redefines shutdown process for window to close equipment."""
        super(QMainWindow, self).closeEvent(*args, **kwargs)
        self.timer.stop()
        self.timer.disconnect()
        self.disconnect()  # add shutdown process when window closed

    def disconnect(self):
        """Run shutdown sequence then disconnect communications."""
        self.shut_down()
        if self.gas_Status.isChecked():
            self.gas_controller.disconnect()

        if self.heater_Status.isChecked():
            self.heater.disconnect()

        if self.gc_Status.isChecked():
            self.gc_connector.disconnect()

    def shut_down(self):
        """Run shutdown method on each connected piece of equipment."""
        print('Shutting Down Equipment')
        if self.gas_Status.isChecked():
            self.gas_controller.shut_down()

        if self.heater_Status.isChecked():
            self.heater.shut_down()

        if self.diode_Status.isChecked():
            self.laser_controller.shut_down()

    def emergency_stop(self):
        """Cancel active threads, call self.shut_down()."""
        # self.threadpool.clear()
        # self.threadpool.cancel()
        self.threadpool.cancel(self.run_study_thread)
        self.threadpool.cancel(self.manual_ctrl_thread)
        # self.threadpool.disconnect()
        self.shut_down()

class EmittingStream():
    """
    Capture console print statements and broadcast within the GUI.

    Redefine sys.stdout, which typically handles all print statements.
    We rewrite the write method to also write to a given Qtextedit.

    Arguments
    ---------
    textedit : QLineEdit
        This should be a line edit box you want to populate w/ print statement.
    """
    def __init__(self, textedit):
        self.textbox = textedit
        self.terminal = sys.stdout
        sys.stdout = self

    def write(self, text):
        """Send captured text to sys.stdout and textedit."""
        self.terminal.write(str(text))
        cursor = self.textbox.textCursor()
        cursor.movePosition(QTextCursor.End)
        cursor.insertText(text)
        self.textbox.setTextCursor(cursor)
        self.textbox.ensureCursorVisible()

    def flush(self):
        """Force print to terminal without buffering"""
        self.terminal.flush()

class WorkerSignal(QObject):
    """Provide signals for interacting w/ worker threads."""
    finished = pyqtSignal()

class Worker(QRunnable):
    """
    Worker thread.

    Inherits from QRunnable to handle
    worker thread setup, signals and wrap-up.

    Parameters
    ----------
    callback : function
        The function callback to run on this worker thread. Supplied args and
        kwargs will be passed through to the runner.

    args:
        Arguments to pass to the callback function
    kwargs:
        Keywords to pass to the callback function

    References
    ----------
    [1] (https://www.pythonguis.com/tutorials/
         multithreading-pyqt-applications-qthreadpool/)

    """

    def __init__(self, fn, *args, **kwargs):
        super(Worker, self).__init__()
        # Store constructor arguments (re-used for processing)
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self.signal = WorkerSignal()

    @pyqtSlot()
    def run(self):
        """Initialise the runner function with passed args, kwargs."""
        self.fn(*self.args, **self.kwargs)
        self.signal.finished.emit()


def setup_style(app):
    """Pull in the .qss sheet for GUI style."""
    script_dir = os.path.dirname(__file__)  # <-- absolute dir the script is in
    rel_path = r"gui_components\style_guide\Adaptic\Adaptic_v3.qss"
    abs_file_path = os.path.join(script_dir, rel_path)
    file = open(abs_file_path, 'r')
    with file:
        qss = file.read()
        app.setStyleSheet(qss)


if __name__ == "__main__":
    # Main
    plt.close('all')
    update_flag = False
    app = QApplication(sys.argv)
    window = MainWindow()
    window.show()
    setup_style(app)
    sys.exit(app.exec_())



"""
Subpackage used for analyzing data collected by catalight automation tools.

Plotting, user dialogs, and scripts contained within. Note: any module prefixed
'run\\_' can be executed directly to utilize the a simple gui dialog or used by
calling module.main() function.
"""
from . import tools  # noqa
from . import plotting  # noqa
from . import user_inputs  # noqa



"""Control Diode laser via DAQ board."""
import datetime as dt
import os
import re
import time
from ctypes import POINTER, cast
from threading import Timer

import numpy as np
import pyttsx3
from comtypes import CLSCTX_ALL
from mcculw import ul
from mcculw.device_info import DaqDeviceInfo
from pycaw.pycaw import AudioUtilities, IAudioEndpointVolume

# Sets path when file is imported
package_dir = os.path.dirname(os.path.abspath(__file__))
calibration_path = os.path.join(package_dir, 'diode_calibration.txt')


class Diode_Laser():
    """
    Virtual instance of diode laser.

    Virtual object for interfacing with Thorlabs diode driver system via an
    MCC DAQ device. The DAQ board should be connected to the current control
    system via BNC cables. A voltage is supplied by the DAQ board to send a
    current from the current controller to the diode driver, thus emitting
    a specified power from the laser. The user must perform a calibration of
    the system prior to using this class!
    """

    def __init__(self):
        # Set public attr
        self.is_busy = False  #: Used to block access from multiple threads
        self.board_num = 0  #: Location of DAQ in "instacal" software.
        self.memhandle = None
        self.channel = 0
        self.dev_id_list = []
        self._calibration = [0, 0]

        # Set non-public attr
        self._I_max = 2000  #: (mA) Max current of current controller
        self._k_mod = self._I_max / 10  #: (mA/V)
        self._P_set = 0

        # Initiate DAQ
        self._daq_dev_info = DaqDeviceInfo(self.board_num)
        self._ao_info = self._daq_dev_info.get_ao_info()

        if self._ao_info.is_supported:
            self._ao_range = self._ao_info.supported_ranges[0]
        else:
            print('Warning: Output not supported by DAQ')

        self._ai_info = self._daq_dev_info.get_ai_info()

        if self._ai_info.is_supported:
            self._ai_range = self._ai_info.supported_ranges[0]
        else:
            print('Warning: Input not supported by DAQ')

        # Initialize equipment
        print('Active DAQ device: ', self._daq_dev_info.product_name, ' (',
              self._daq_dev_info.unique_id, ')\n', sep='')
        self.read_calibration()

        # Initiate a voice control object to send alert messages
        self.voice_control = pyttsx3.init()
        self.voice_control.setProperty('volume', 1.0)
        rate = self.voice_control.getProperty('rate')
        self.voice_control.setProperty('rate', rate + 1)

        # Initiate control over speakers
        devices = AudioUtilities.GetSpeakers()
        interface = devices.Activate(IAudioEndpointVolume._iid_,
                                     CLSCTX_ALL, None)
        self.volume_control = cast(interface, POINTER(IAudioEndpointVolume))

        # Turn power to zero on initialization
        self.set_power(0)

    # Read Only Attributes
    I_max = property(lambda self: self._I_max)  #: (mA) Max current
    k_mod = property(lambda self: self._k_mod)  #: (mA/V) Conversion Factor
    daq_dev_info = property(lambda self: self._daq_dev_info)  #: DAQ info
    ao_info = property(lambda self: self._ao_info)  #: Analog output info
    ao_range = property(lambda self: self._ao_range)  #: Analog output range
    ai_info = property(lambda self: self._ai_info)  #: Analog input info
    ai_range = property(lambda self: self._ai_range)  #: Analog input range
    P_set = property(lambda self: self._P_set)  #: Current laser setpoint

    def set_power(self, P_set):
        """
        Send signal to DAQ to reach desired P_set value based on calibration.

        The necessary current is sent based on a externally performed
        calibration.
        * Raises speaker volume
        * Sends voice warning
        * Prints time to complete ramp
        * ramps from current setpoint to P_set at 650 mA/min to avoid equipment damage
        * Redefined _P_set along the way, printing status at each step
        * Calls print_output() on completion


        Parameters
        ----------
        P_set : int or float
            Desired laser output power in mW
        """
        # TODO put check on max power
        self.voice_control.setProperty('volume', 1.0)
        # TODO can i make this try to be int
        self.voice_control.say('Warning: Setting power to %6.2f milliwatts' % P_set)
        self.voice_control.runAndWait()
        self.voice_control.stop()
        # speak('Warning: Setting power to' + str(P_set) + 'milliwatts')
        I_set = self.P_to_I(P_set)  # (mA) Based on calibration
        I_start = self.get_output_current()
        # TODO can i round this to 0 if negative?
        P_start = self.I_to_P(I_start)
        if P_start < 0:
            I_start = self.P_to_I(0)
        refresh_rate = 20  # 1/min
        ramp_time = (I_set - I_start) / 650  # [min] - spans 1300mA in 2 min
        setpoints = np.linspace(I_start, I_set,
                                abs(int(ramp_time * refresh_rate)))
        setpoints = np.append(setpoints, I_set)
        if P_set != 0:
            print('ramp time = %6.4f minutes' % ramp_time)

        while self.is_busy:
            time.sleep(0)
        self.is_busy = True

        for I in setpoints: # noqa I==current
            # Ramps the current slowly
            Vout = I / self._k_mod  # (V) Voltage output set point
            if (P_set == 0) and self._ao_info.is_supported:
                Vout = 0
                # Convert to 16bit
                Vout_value = ul.from_eng_units(self.board_num, self._ao_range, Vout)
                # Send signal to DAQ Board
                ul.a_out(self.board_num, 0, self._ao_range, Vout_value)
                break

            elif (P_set != 0) and self._ao_info.is_supported:
                # Convert to 16bit
                Vout_value = ul.from_eng_units(self.board_num, self._ao_range, Vout)
                # Send signal to DAQ Board
                ul.a_out(self.board_num, 0, self._ao_range, Vout_value)
                time.sleep(60 / refresh_rate)  # wait
                self._P_set = self.I_to_P(I)
                print('Set Point = %7.2f mW / %7.2f mA' % (self.P_set, I))

            else:
                print('DAQ Write Not Supported')

        self.is_busy = False
        self._P_set = P_set
        print('\n', time.ctime())
        self.print_output()
        print('Set Point = %7.2f mW / %7.2f mA \n' % (self.P_set, I_set))

    def get_output_current(self):
        """
        Convert DAQ voltage to current measurement.

        Returns
        -------
        float or int
            Current measured by DAQ

        """
        while self.is_busy:
            time.sleep(0)
        self.is_busy = True

        if self._ai_info.is_supported:
            # Get input value into DAQ
            Vin_value = ul.a_in(self.board_num, self.channel, self._ai_range)
            Vin_eng_units_value = ul.to_eng_units(self.board_num,
                                                  self._ai_range, Vin_value)
            # Convert to relevant output numbers
            V = Vin_eng_units_value
            I = round(V * self._k_mod, 3) #noqa I==current
        else:
            print('DAQ Read Not Supported')
            I = 0 #noqa I==current

        self.is_busy = False
        return(abs(I))

    def get_output_power(self):
        """
        Get output power based on current measured and saved calibration.

        Returns
        -------
        P : float or int
            Power [mW] rounded to 3 decimal points
        """
        I = self.get_output_current() #noqa I==current
        P = round(self.I_to_P(I), 3)
        return(P)

    def print_output(self):
        """Print the output current and power to console."""
        I = self.get_output_current() #noqa I==current
        P = self.get_output_power()
        print('Measured Laser output = %7.2f mW / %7.2f mA' % (P, I))

    def I_to_P(self, I): #noqa I==current
        """
        converts a current to power based on read calibration

        Parameters
        ----------
        I : current you'd like to convert [mA]

        Returns
        -------
        P : Power [mW]

        """
        m = self._calibration[0]
        b = self._calibration[1]
        P = I * m + b
        return(P)

    def P_to_I(self, P):
        """
        Convert a power to current based on read calibration.

        Parameters
        ----------
        P : int or float
            Power you'd like to convert [mW]

        Returns
        -------
        I : float
            Equivalent current [mA]

        """
        m = self._calibration[0]
        b = self._calibration[1]
        I = (P-b)/m  # (mA) Based on calibration #noqa I==current
        return(I)

    def shut_down(self):
        """Set power of laser to 0 by setting DAQ Voltage to 0."""
        while self.is_busy:
            time.sleep(0)
        self.is_busy = True
        Vout = 0  # (V) Voltage output set point
        # Convert to 16bit
        Vout_value = ul.from_eng_units(self.board_num, self._ao_range, Vout)
        # Send signal to DAQ Board
        ul.a_out(self.board_num, 0, self._ao_range, Vout_value)
        self.is_busy = False

    def update_calibration(self, slope, intercept):
        """
        Write calibration info to file.

        Calibration file path is defined at module level on import.
        Also saves date of calibration. Overwrites previous calibration!

        Parameters
        ----------
        slope : float
            Slope of linear calibration (mW/mA)
        intercept : float
            Y intercept of linear calibration (mA)
        """
        # open and write to calibration file
        with open(calibration_path, 'r+') as old_cal_file:
            new_cal_file = []

            for line in old_cal_file:  # read values after '=' line by line
                if re.search('m = ', line):
                    line = ('m = ' + str(slope) + ' \n')
                elif re.search('b = ', line):
                    line = ('b = ' + str(intercept) + ' \n')
                elif re.search('date = ', line):
                    line = ('date = ' + dt.date.today().strftime('%Y-%m-%d') + '\n')

                new_cal_file += line

            old_cal_file.seek(0)  # Starting from beginning line
            old_cal_file.writelines(new_cal_file)
        self.read_calibration()

    def read_calibration(self):
        """
        Update calibration attr based on module level calibration file.

        Also prints out calibration date and values to console.
        """
        with open(calibration_path, 'r') as calibration:

            for line in calibration:  # read values after '=' line by line
                if re.search('m = ', line):
                    self._calibration[0] = float(
                        line.split('=')[-1].strip(' \n'))
                elif re.search('b = ', line):
                    self._calibration[1] = float(
                        line.split('=')[-1].strip(' \n'))
                elif re.search('date = ', line):
                    print('Last laser calibration was:')
                    print(line)
        print(f'Power = {self._calibration[0]}*Current'
              f'+ {self._calibration[1]}\n')

    def time_warning(self, time_left):
        """
        Play audio warning that laser will engage in time_left minutes.

        Parameters
        ----------
        time_left : int or float
            (min) Time to read out until setting laser
        """
        # Consider upgrading this to use asyncio or threading.Timer and have
        # the code put out 5 4 3 2 1 minute warnings on a separate thread
        # Unmutes and sets Vol in dB -0.0 is 100%
        self.volume_control.SetMute(0, None)
        self.voice_control.say(f'Warning: Diode laser will automatically'
                               f'engage in {time_left} minutes')
        self.voice_control.runAndWait()

    def set_current(self, I_set):
        """
        Set current output of controller.

        Use this when running calibration. Reads warning messages when
        changing power

        Parameters
        ----------
        I_set : int or float
            Current setpoint to send to controller (mA)
        """
        Vout = I_set / self._k_mod  # (V) Voltage output set point
        print('Current set to %.2f\nVoltage set to %.2f' % (I_set, Vout))
        while self.is_busy:
            time.sleep(0)
        self.is_busy = True
        # Convert to 16bit
        Vout_value = ul.from_eng_units(self.board_num, self._ao_range, Vout)

        # Send signal to DAQ Board
        ul.a_out(self.board_num, 0, self._ao_range, Vout_value)
        Vin_value = ul.a_in(self.board_num, self.channel, self._ai_range)
        Vin_eng_units_value = ul.to_eng_units(self.board_num,
                                              self._ai_range, Vin_value)
        self.is_busy = False

        self.print_output()
        print(time.ctime())
        # Unmutes and sets Vol in dB -0.0 is 100%
        self.volume_control.SetMute(0, None)
        self.volume_control.SetMasterVolumeLevel(-2.0, None)
        self.voice_control.say('Warning: Setting current to %6.2f milliamps' % I_set)
        self.voice_control.runAndWait()

    def start_logger(self, log_frequency=0.1, save_path=None):
        """
        Start the data log function to record the laser set point.

        Records constantly until stop_logger() is called. Creates self.timer
        and self.save_path when called.

        Parameters
        ----------
        log_frequency : float or int, optional
            (seconds) interval to record data with. The default is 0.1 sec.
        save_path : str, optional
            Full tile path to save data to. If None, saves in module directory
            with file name 'YYYYMMDDlaser_log.txt'. Appends int to end of file
            name if file name already exists.
        """
        if save_path is None:  # if savepath is unspecified, saves in cwd
            save_path = os.path.join(os.getcwd(),
                                     dt.date.today().strftime('%Y%m%d') + 'laser_log.txt')
        while os.path.isfile(save_path):  # check if log w/ default name exists
            m = 1
            save_path = save_path.removesuffix('.txt')
            # checks if save_path has number at end
            if re.findall(r'\d+$', save_path):
                m += int(re.findall(r'\d+$', save_path)[-1])
            # append number and .txt to filename,
            # then check if name still exists
            save_path = re.split(r'\d+$', save_path)[0] + str(m)
            save_path = save_path + '.txt'

        self.save_path = save_path  # assign local pathname to class attribute
        # create update thread
        self.timer = RepeatTimer(log_frequency, self.log_power)
        self.timer.start()

    def log_power(self):
        """Append (date and current power setpoint) to the log at save_path."""
        with open(self.save_path, 'a') as output_log:
            entry = ('%s, %6.2f \n' % (dt.datetime.now(), self.P_set))
            output_log.write(entry)

    def stop_logger(self):
        """Cancel timer initialized w/ start_logger() and delete it."""
        if self.timer:
            self.timer.cancel()
            del self.timer
        else:
            print('No Timer')


class RepeatTimer(Timer):
    """
    Subclass of threading.Timer. Runs at interval until cancelled.

    Parameters
    ----------
    interval : float or int
        Time interval in seconds to call function.
    function : func
        Functions to call
    args
        Arguments of function.
    kwargs
        Key word arguments of function.
    """

    def run(self):
        """Run function until finished supplying args and kwargs."""
        while not self.finished.wait(self.interval):
            self.function(*self.args, **self.kwargs)


if __name__ == "__main__":
    laser_controller = Diode_Laser()
    laser_controller.start_logger()
    laser_controller.time_warning(1)
    laser_controller.print_output()
    laser_controller.stop_logger()
    laser_controller.shut_down()
